@@ -5,7 +5,7 @@
 		- unpaper
 		- hocr2pdf (from ExactImage)
 	
-	(C) 2010-2013 Tobias Elze, modified for tesseract Heinrich Schwietering 2012
+	(C) 2010-2016 Tobias Elze, modified for tesseract Heinrich Schwietering 2012
 	patch for -rgb option contributed by James Cort
 *)
 
@@ -16,8 +16,10 @@ let unpaper = ref "unpaper";;
 let identify = ref "identify";;
 let convert = ref "convert";;
 let tesseract = ref "tesseract";;
-let gs = ref "gs";;
+let pdfinfo = ref "pdfinfo";;
+let pdfunite = ref "pdfunite";;
 let hocr2pdf = ref "hocr2pdf";;
+let gs = ref "gs";;
 
 (*global flags:*)
 let verbose = ref false;;
@@ -29,9 +31,17 @@ let pr s =
 ;;
 
 (*execute command cmd and print it's invocation line (if verbose is set):*)
-let run cmd =
+let run ?(crash=true) cmd =
 	if !verbose then pr cmd;
-	ignore (Sys.command cmd);
+	if Sys.command cmd <> 0 then
+	(
+		prerr_endline ("ERROR: Command \"" ^ cmd ^ "\" failed. ");
+		if crash  then
+		(
+			prerr_endline ("Terminating pdfsandwich. All temporary files are kept.");
+			exit 2;
+		)
+	)
 ;;
 
 (*check if binary bin exists (in search path):*)
@@ -47,20 +57,24 @@ let check_for_binary bin =
 		^ ". Make sure this program exists and can be found in your search path.\nUse command line options to specify a custom binary.")
 ;;
 
-(*return number of pages of a PDF file; needs gs:*)
+(*return number of pages of a PDF file; needs pdfinfo:*)
 let number_of_pages filename =
 	try
-	let in_ch = Unix.open_process_in
-		("printf \"/PDFfile PDFname (r) file def\n\n/PageCountString 255 string\n\ndef\nsystemdict /.setsafe known { .setsafe } if\n/.show.stdout { (%%stdout) (w) file } bind def\n/puts { .show.stdout exch writestring } bind def\nGS_PDF_ProcSet begin\npdfdict begin\nPDFfile\npdfopen begin\n/FirstPage where { pop } { /FirstPage 1 def } ifelse\n/LastPage where { pop } { /LastPage pdfpagecount def } ifelse\n() puts\nLastPage FirstPage sub 1 add\n\nPageCountString cvs puts\n(\\\\\\\\n) print\nquit \n\""
-		^ "|" ^ !gs ^ " -q -sNODISPLAY -sPDFname=\""
-		^ filename ^ "\" -")
+	let in_ch = Unix.open_process_in (!pdfinfo ^ " " ^ filename) in
+	let rec read_pdfinfo () =
+		let s = input_line in_ch in
+		if Str.string_match (Str.regexp "Pages:[^0-9]+\\([0-9]+\\)") s 0 then
+		(
+			ignore (Unix.close_process_in in_ch);
+			int_of_string (Str.matched_group 1 s);
+		)
+		else read_pdfinfo ()
 	in
-	let pagestring = input_line in_ch in
-	ignore (Unix.close_process_in in_ch);
-	int_of_string pagestring;
-	with _ -> failwith ("Error: Could not determine number of pages of file " ^ filename)
+	read_pdfinfo ()
+	with _ -> failwith ("Error: pdfinfo could not determine number of pages. Check the pdf input file.\n")
 ;;
 
+
 (*return number of CPUs:*)
 let number_of_cpus () =
 	(*different code for different platforms;
@@ -74,8 +88,9 @@ let number_of_cpus () =
 			let os = input_line p in
 			ignore (Unix.close_process_in p);
 			match os with 
-			| "Darwin" -> "sysctl -n hw.ncpu"
-			| _ -> "cat /proc/cpuinfo | grep processor | awk '{a++} END {print a}'" (*Linux*)
+			| "Linux" -> "cat /proc/cpuinfo | grep processor | awk '{a++} END {print a}'" 
+			| _ -> "sysctl -n hw.ncpu"	(*works (at least) for Darwin, OpenBSD, and FreeBSD*)
+			
 		in
 		let in_ch = Unix.open_process_in scriptstring in
 		let numstr = input_line in_ch in
@@ -93,6 +108,7 @@ let process_ocr
 				first_page last_page 
 				resolution 
 				rgb 
+				gray
 				nthreads 
 				language 
 				convertopts 
@@ -111,64 +127,78 @@ let process_ocr
 		pr ("\nParallel processing with " ^ (string_of_int nthreads) ^ " threads started.\nProcessing page order may differ from original page order.\n");
 	
 	let process_page (curr_page, pdfname) =
-		let tmppicfile = Filename.temp_file "pdfsandwich" (if rgb then ".ppm" else ".pbm") in
+		let tmppicfile = Filename.temp_file "pdfsandwich" (if rgb then ".ppm" else if gray then ".pgm" else ".pbm") in
+		let tmptessinpfile = Filename.temp_file "pdfsandwich" ".tif" in
 		let tmpocrfile = Filename.temp_file "pdfsandwich" "" in
 		let tmpcolfigfile = Filename.temp_file "pdfsandwich" "_col.png" in
-		let tmpdecreased_infile = Filename.temp_file "pdfsandwich" "_rescaled.pdf" in
-		let tmpunpaperfile = Filename.temp_file "pdfsandwich" ("_unpaper" ^ (if rgb then ".ppm" else ".pbm")) in
+		let tmprescaled_infile = Filename.temp_file "pdfsandwich" "_rescaled.pdf" in
+		let tmpunpaperfile = Filename.temp_file "pdfsandwich" ("_unpaper" ^ (if rgb then ".ppm" else if gray then ".pgm" else ".pbm")) in
 		if not !quiet then
 			Printf.printf "Processing page %i.\n" curr_page;
 		flush_all ();
 		(*get original height and width:*)
+		let (orig_height, orig_width) =
+			try
+			pr (!identify ^ " -format \"%w\\n%h\\n\" " ^ " \"" ^ infile ^ "[" ^ (string_of_int (curr_page-1)) ^ "]\" ");
+			let in_ch = Unix.open_process_in
+				(!identify ^ " -format \"%w\\n%h\\n\" " ^ " \"" ^ infile ^ "[" ^ (string_of_int (curr_page-1)) ^ "]\" ");
+			in
+			let w = input_line in_ch 
+			and h = input_line in_ch  in
+			ignore (Unix.close_process_in in_ch);
+			(int_of_string h, int_of_string w);
+			with _ -> (
+				prerr_endline "Warning: could not determine page size; defaulting to A4.";
+				(842,595) (*defaults to A4*))
+		in
 		let (height, width) = match page_width_height with
-			| None ->
-			(	
-				try
-				pr (!identify ^ " -format \"%w\\n%h\\n\" " ^ " \"" ^ infile ^ "[" ^ (string_of_int (curr_page-1)) ^ "]\" ");
-				let in_ch = Unix.open_process_in
-					(!identify ^ " -format \"%w\\n%h\\n\" " ^ " \"" ^ infile ^ "[" ^ (string_of_int (curr_page-1)) ^ "]\" ");
-				in
-				let w = input_line in_ch 
-				and h = input_line in_ch  in
-				ignore (Unix.close_process_in in_ch);
-				(int_of_string h, int_of_string w);
-				with _ -> (
-					prerr_endline "Warning: could not determine page size; defaulting to A4.";
-					(842,595) (*defaults to A4*))
-			)
+			| None -> (orig_height, orig_width)
 			| Some x -> x
 		in
-		(*downscaling if resolution too large:*)
+		(*downscaling if resolution too large (requires gs):*)
 		let convert_infile =
 			let fw = float_of_int width
 			and fh = float_of_int height
 			and fm = float_of_int maxpixels in
 			let pixels = (((float_of_int resolution) /. 72.) ** 2.) *. fw *. fh in
-			if pixels > fm then
+			
+			let (new_height, new_width) =
+				if pixels > fm then
+				(
+					let mpix = fm /. (((float_of_int resolution) /. 72.) ** 2.) in
+					let k = fw /. fh in
+					let new_width = int_of_float (sqrt (k *. mpix)) in
+					let new_height = int_of_float ((sqrt mpix) /. (sqrt k)) in
+					Printf.eprintf 
+						"\n\nWARNING: page size (%ix%i) of page %i together with resolution %i yields very large file which exceeds parameter maxpixels. Most probably, the input file was accidentally generated in an inappropriately hight resolution.\nThe input file is scaled down to %ix%i pixels instead.\nIf such a large input file is really required, set the command line option -maxpixels greater than %.0f instead.\n\n" 
+						width height curr_page resolution new_width new_height pixels;
+					flush_all ();
+					(new_height, new_width)
+				)
+				else (height, width)
+				
+			in
+			
+			if (new_height <> orig_height || new_width <> orig_width) then
 			(
-				let mpix = fm /. (((float_of_int resolution) /. 72.) ** 2.) in
-				let k = fw /. fh in
-				let new_width = int_of_float (sqrt (k *. mpix)) in
-				let new_height = int_of_float ((sqrt mpix) /. (sqrt k)) in
-				Printf.eprintf 
-					"\n\nWARNING: page size (%ix%i) of page %i together with resolution %i yields very large file which exceeds parameter maxpixels. Most probably, the input file was accidentally generated in an inappropriately hight resolution.\nThe input file is scaled down to %ix%i pixels instead.\nIf such a large input file is really required, set the command line option -maxpixels greater than %.0f instead.\n\n" 
-					width height curr_page resolution new_width new_height pixels;
-				flush_all ();
 				run (
 					Printf.sprintf 
 						"%s  -q -dNOPAUSE -dBATCH -sDEVICE=pdfwrite -dFirstPage=%i -dLastPage=%i -dDEVICEWIDTHPOINTS=%i -dDEVICEHEIGHTPOINTS=%i -dPDFFitPage -o %s %s" 
-						!gs curr_page curr_page new_width new_height tmpdecreased_infile infile);
-				tmpdecreased_infile;
+						!gs curr_page curr_page new_width new_height tmprescaled_infile infile);
+				tmprescaled_infile;
 			)
 			else
 				(infile ^ "[" ^ (string_of_int (curr_page-1)) ^ "]")
 		in
 		let convoptstmp =
-			if rgb then " -depth 8 -background white -flatten +matte -density " else " -type Bilevel  -density "
+			if rgb then 
+				" -depth 8 -background white -flatten -alpha Off -density " 
+			else 
+				if gray then " -colorspace gray -depth 8 -background white -flatten -alpha Off -density " else " -type Bilevel -density "
 		in
-		run (!convert ^ " " ^ convertopts ^ convoptstmp ^ (Printf.sprintf "%ix%i" resolution resolution) ^ " \"" ^ convert_infile ^ "\" " ^ tmppicfile);
+		run (!convert ^ " -units PixelsPerInch " ^ convertopts ^ convoptstmp ^ (Printf.sprintf "%ix%i " resolution resolution)  ^ " \"" ^ convert_infile ^ "\" " ^  tmppicfile);
 		let tessout = if not !verbose then ">/dev/null 2>&1" else "" in
-		let tessinputfile = 
+		let preproc_output = 
 			if preprocess then
 			(
 				run (!unpaper ^ " --overwrite " ^ unpaperopts ^ " " ^ tmppicfile ^ " " ^ tmpunpaperfile ^ tessout);
@@ -176,19 +206,24 @@ let process_ocr
 			)
 			else tmppicfile
 		in
+		(*convert preprocessing output file to tif in order to ensure correct resolution and size:*)
+		let tessinputfile = 
+			run (!convert ^ " -units PixelsPerInch -density " ^ (Printf.sprintf "%ix%i" resolution resolution) ^ " " ^ preproc_output ^ " " ^ tmptessinpfile );
+			tmptessinpfile
+		in
 		
 		(*test if tesseract can output pdf files:*)
 		run (!tesseract ^ " " ^ tessinputfile ^ tessout ^ " " ^ tmpocrfile ^ " " ^ tessopts ^ " -l " ^ language ^ " pdf ");
 		
 		if (not enforcehocr2pdf) && Sys.file_exists (tmpocrfile ^ ".pdf") then
 		(
-			(* resetting page size: *)
-			run (Printf.sprintf "%s -q -dNOPAUSE -dBATCH -sDEVICE=pdfwrite -dDEVICEWIDTHPOINTS=%i -dDEVICEHEIGHTPOINTS=%i -dPDFFitPage -o %s %s.pdf" !gs width height pdfname tmpocrfile);
+			if !verbose then pr ("OCR pdf generated. Renaming output file to " ^ pdfname ^ "\n");
+			Unix.rename (tmpocrfile ^ ".pdf") pdfname;
 		)
 		else
 		(
 			if not !quiet then
-				print_endline "Version of tesseract is prior to 3.03 and cannot output pdf yet. Using hocr2pdf instead.";
+				print_endline "Tesseract was unable to produce a pdf ouput file. Possibly, version of tesseract is prior to 3.03 and cannot output pdf yet. Using hocr2pdf instead.";
 			run (!tesseract ^ " " ^ tessinputfile ^ tessout ^ " " ^ tmpocrfile ^ " " ^ tessopts ^ " -l " ^ language ^ " hocr ");
 			let hocrinputfile = if Sys.file_exists (tmpocrfile ^ ".html") then tmpocrfile ^ ".html" else tmpocrfile ^ ".hocr" in
 			run (!hocr2pdf ^ hocropts ^ " -r " ^ hocr_resolution ^ " -i " ^ tessinputfile ^ " -o " ^ pdfname ^ "<" ^ hocrinputfile);
@@ -199,7 +234,8 @@ let process_ocr
 		if not debug then
 		(
 			rm_if_exists (tmpocrfile ^ ".pdf");
-			rm_if_exists tmpdecreased_infile;
+			rm_if_exists tmprescaled_infile;
+			rm_if_exists tmptessinpfile;
 			rm_if_exists tmppicfile;
 			rm_if_exists tmpocrfile;
 			rm_if_exists tmpcolfigfile;
@@ -232,8 +268,22 @@ let process_ocr
 	let pdffilenamelist = List.map snd (Array.to_list tmppdf_arr) in
 	let pdfliststring = String.concat " " pdffilenamelist in
 	pr ("OCR done. Writing \"" ^ outfile ^ "\"");
-	run (!gs ^ " -q -dNOPAUSE -dBATCH -sDEVICE=pdfwrite -sOutputFile=\"" ^ outfile ^ "\" " ^ pdfliststring);
-	if (not debug) then List.iter Sys.remove pdffilenamelist;
+	
+	
+	let tmpoutfile = 
+		if List.length pdffilenamelist > 1 then
+		(
+			let tout = Filename.temp_file "pdfsandwich_output" ".pdf" in
+			run (!pdfunite ^ " " ^ pdfliststring ^ " " ^ tout);
+			if (not debug) then List.iter Sys.remove pdffilenamelist;
+			tout
+		)
+		else List.hd pdffilenamelist
+	in
+	
+	try
+	Unix.rename tmpoutfile outfile;
+	with _ -> run("mv \"" ^ tmpoutfile ^ "\" \"" ^ outfile ^ "\"");
 ;;
 
 let main () =
@@ -245,6 +295,7 @@ let main () =
 	let maxpixels = ref 17415167 in
 	let lang = ref "eng" in
 	let rgb = ref false in
+	let gray = ref false in
 	let preprocess = ref true in
 	let unpaperopts = ref "" in
 	let layout = ref "none" in
@@ -292,8 +343,9 @@ let main () =
 		("-debug", Arg.Set debug, "\t keep all temporary files in /tmp (for debugging)");
 		("-enforcehocr2pdf", Arg.Set enforcehocr2pdf, "\t use hocr2pdf even if tesseract >= 3.03");
 		("-first_page", Arg.Set_int first_page, "\t -first_page number : number of page to start OCR from (default: 1)");
-		("-grayfilter", Arg.Set grayfilter, "\t\t enable unpaper's gray filter; further options can be set by -unpo");
-		("-gs", Arg.Set_string gs, "\t\t -gs filename : name of gs binary (default: gs)");
+		("-grayfilter", Arg.Set grayfilter, "\t enable unpaper's gray filter; further options can be set by -unpo");
+		("-gray", Arg.Set gray, "\t use grayscale for images (default: black and white);\n\t\t  will be overridden by use of rgb");
+		("-gs", Arg.Set_string gs, "\t\t -gs filename : name of gs binary (default: gs); optional, only required for resizing");
 		("-hocr2pdf", Arg.Set_string hocr2pdf, "\t -hocr2pdf filename : name of hocr2pdf binary (default: hocr2pdf);\n\t\t  ignored for tesseract >= 3.03 unless option -enforcehocr2pdf is set");
 		("-hoo", Arg.String (fun s -> set_hocr_opts s()), "\t\t -hoo options : additional hocr2pdf options; make sure to quote");
 		("-identify", Arg.Set_string identify, "\t -identify filename : name of identify binary (default: identify)");
@@ -306,7 +358,9 @@ let main () =
 		("-nopreproc", Arg.Clear preprocess, "\t do not preprocess with unpaper");
 		("-nthreads", Arg.Set_int nthreads, "\t -nthreads number : number of parallel threads (default: guessed number of CPUs; if guessing fails: 1)");
 		("-o", Arg.Set_string outputfile, "\t\t -o filename : output file; default: inputfile_ocr.pdf (if extension is different\n\t\t  from .pdf, original extension is kept)");
-		("-pagesize", Arg.Set_string pagesize, "\t -pagesize { original | NUMxNUM } : set page size of output pdf\n\t\t  original: same as input file (default)\n\t\t  NUMxNUM: width x height in pixel (e.g. for A4: -pagesize 595x842)");
+		("-pagesize", Arg.Set_string pagesize, "\t -pagesize { original | NUMxNUM } : set page size of output pdf (requires ghostscript)\n\t\t  original: same as input file (default)\n\t\t  NUMxNUM: width x height in pixel (e.g. for A4: -pagesize 595x842)");
+		("-pdfinfo", Arg.Set_string pdfinfo, "\t -pdfinfo filename : name of pdfinfo binary (default: pdfinfo)");
+		("-pdfunite", Arg.Set_string pdfunite, "\t -pdfunite filename : name of pdfunite binary (default: pdfunite)");
 		("-resolution", Arg.Set_int resolution, "\t -resolution NUM : resolution (dpi) used for OCR (default: 300)");
 		("-rgb", Arg.Set rgb, "\t\t use RGB color space for images (default: black and white);\n\t\t  use with care: causes problems with some color spaces");
 		("-sloppy_text", Arg.Unit (set_hocr_opts "-s"), "\t sloppily place text, group words, do not draw single glyphs;\n\t\t  ignored for tesseract >= 3.03 unless option -enforcehocr2pdf is set");
@@ -320,7 +374,7 @@ let main () =
 	] in
 	
 	Arg.parse speclist (fun s -> arg_filename := s) "USAGE: pdfsandwich [options] inputfile.pdf\n\nOptions:";
-	List.iter check_for_binary [!convert; !tesseract; !gs];
+	List.iter check_for_binary [!convert; !tesseract; !gs; !pdfunite];
 	if !enforcehocr2pdf then check_for_binary !hocr2pdf;
 	if !preprocess then check_for_binary !unpaper;
 	if !outputfile = "" then 
@@ -335,13 +389,15 @@ let main () =
 	(
 		let check_version progstring opt = 
 			print_endline ("Checking for " ^ progstring ^ ":");
-			run (progstring ^ " " ^ opt);
+			run ~crash:false (progstring ^ " " ^ opt);
 		in
 		check_version !convert "-version";
-		if !preprocess then check_version !unpaper "-version";
+		if !preprocess then check_version !unpaper "-V";
 		check_version !tesseract "-v";
 		if !enforcehocr2pdf then check_version !hocr2pdf "-h";
 		check_version !gs "-v";
+		check_version !pdfinfo "-v";
+		check_version !pdfunite "-v";
 	);
 	
 	if not !grayfilter then set_unp_opts "--no-grayfilter" ();
@@ -368,7 +424,14 @@ let main () =
 	
 	pr ("Input file: \"" ^ !arg_filename ^ "\"");
 	pr ("Output file: \"" ^ !outputfile ^ "\"");
-	let npages = number_of_pages !arg_filename in
+	(*use different filename for the input file in case the original input file contains special characters:*)
+	let internal_input_filename = Filename.temp_file "pdfsandwich_inputfile" ".pdf" in
+	Sys.remove internal_input_filename;
+	(*link original input file to internal_input_filename:*)
+	let full_path_input_file = if Filename.is_relative !arg_filename then ((Sys.getcwd ()) ^ "/" ^ !arg_filename) else !arg_filename in
+	Unix.symlink full_path_input_file internal_input_filename;
+	
+	let npages = number_of_pages internal_input_filename in
 	pr ("Number of pages in inputfile: " ^ (string_of_int npages));
 	if (!first_page < 1 || !first_page > npages) then failwith ("Value " ^ (string_of_int !first_page) ^ " is invalid as first_page.");
 	if !last_page < 1 then last_page:= npages;
@@ -383,11 +446,12 @@ let main () =
 	);
 	
 	process_ocr 
-		!arg_filename 
+		internal_input_filename 
 		!outputfile 
 		!first_page !last_page 
 		!resolution 
 		!rgb 
+		!gray
 		!nthreads 
 		!lang 
 		!convertopts !tessopts !hocropts 
@@ -396,7 +460,8 @@ let main () =
 		!enforcehocr2pdf 
 		page_width_height
 		!maxpixels;
-	pr "\nDone.";
+	Unix.unlink internal_input_filename;
+	pr ("\n" ^ !outputfile ^ " generated.\n\nDone.");
 ;;
 
 main ();;
